#!/usr/bin/perl

# rip-utils
#
# Copyright (C) 2001-2002 Mikol Graves
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA.

# ---------------------------------------------------------------------------
package tbx::CDDB;
# ---------------------------------------------------------------------------

use strict;

use tbx::Album;

use IO::Socket;
use vars qw( @ISA @EXPORT );

@ISA = qw( Exporter );
@EXPORT = qw( &id );

sub CD_MSF_OFFSET       () { 150 }     # From /usr/include/linux/cdrom.h
sub CD_FRAMES           () { 75 }      # From /usr/include/linux/cdrom.h

sub CDDB_HOST           () { "freedb.freedb.org" }
sub CDDB_PORT           () { 888 }
sub CDDB_MODE           () { "cddb" }

# Globals
my @toc;            # CD table of contents
my $id;             # CD's calculated CDDB ID
my $album;          # CD's tbx::Album object
my $xmcd = '';      # xmcd CD database file
my $categ = '';     # effective CDDB category
my $discid = '';    # effective CDDB ID
my $revision = 0;   # revision number - increment when submitting to cddb
my $dyear = 0;      # xmcd DYEAR field
my $dgenre = '';    # xmcd DGENRE field
my $extd = '';      # xmcd EXTD field
my @extt;           # xmcd EXTTN fields
my $playorder = ''; # xmcd PLAYORDER field

# ---------------------------------------------------------------------------
# Generate a valid CDDB disc ID and store it in the $id global.
# ---------------------------------------------------------------------------
sub id {
  $id = undef;
  my ($ok, $err) = toc();
  
  unless (defined($ok)) {
    return (undef, $err);
  }

  my $tracks = scalar(@toc) - 1;
  my $n = 0;
  my $t = 0;

  for (my $i = 0; $i < $tracks; $i++) {
    $n = $n + sum(($toc[$i]->{m} * 60) + $toc[$i]->{s});
  }

  unless ($#toc > -1) { 
    return (undef, "Couldn't determine CD table of contents." .
            " Maybe there is no disc in the drive.");
  }

  $t = int(($toc[$#toc]->{o} - $toc[0]->{o})/75);

  $id = sprintf("%08x", (($n % 0xff) << 24 | $t << 8 | $tracks));

  return $id;
}

# ---------------------------------------------------------------------------
# m ... minutes
# s ... seconds
# f ... frames
# o ... offset
# ---------------------------------------------------------------------------
sub toc {
  @toc = ();
  my $cdparanoia = `which cdparanoia`;
  chomp($cdparanoia);

  if ($cdparanoia) {
    if (-x $cdparanoia) {
      open(CDPARANOIA, "$cdparanoia -Q 2>&1 |");

      while (<CDPARANOIA>) {
        if (/^\s*\d+\.\s*\d+ \[.*?\]\s*(\d+) \[.*$/) {
          my ($m, $s, $f, $o) = lba_to_msf($1);
          push (@toc, { "m" => $m, "s" => $s, "f"  => $f, "o"  => $o });
        }
        if (/^TOTAL\s*(\d+) \[.*$/) {
          my ($m, $s, $f, $o) = lba_to_msf($1);
          push (@toc, { "m" => $m, "s" => $s, "f"  => $f, "o"  => $o });
        }
      }

      close(CDPARANOIA);

      return 1;
    } else {
      return (undef, "You do not have permission to execute cdparanoia.");
    }
  } else {
    return (undef, "The cdparanoia executable is not in your path.");
  }
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
sub lba_to_msf {
  my $o = shift();
  my ($m, $s, $f, $t);

  $o += CD_MSF_OFFSET;
  $f = $o % CD_FRAMES;

  $t = ($o - $f) / CD_FRAMES;

  $s = $t % 60;
  $m = int($t / 60);

  return ($m, $s, $f, $o);
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
sub sum {
  my $n = shift();
  my $ret = 0;

  while ($n > 0) {
    $ret += ($n % 10);
    $n = int($n / 10);
  }

  return $ret;
}

# FreeDB Client Subroutines -------------------------------------------------
# --> Client to Server
# <-- Server to Client
# ---------------------------------------------------------------------------

my $socket;

# ---------------------------------------------------------------------------
# categ_and_discid will try to determine the CDDB category and disc ID given
# the response from a CDDB server. It's intended to be used with the value(s)
# returned by query().
# ---------------------------------------------------------------------------
sub categ_and_discid {
  my ($str) = @_;
  my ($code, $categ, $discid) = split(/\s/, $str);
  if ($code !~ /^\d\d\d$/) {
    $discid = $categ;
    $categ = $code;
  }
  return ($categ, $discid);
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
sub cddb_up {
  return $socket if $socket;

  $socket = new IO::Socket::INET(PeerAddr => CDDB_HOST,
                                 PeerPort => CDDB_PORT,
                                 Proto    => "tcp",
                                 Type     => SOCK_STREAM);
  unless($socket) {
    return (undef, "Could establish connection with ".CDDB_HOST." ".CDDB_PORT.": $!");
  }

  my $ret = <$socket>;

  unless ($ret =~ /^2\d\d\s+/) {
    return (undef, "Connection refused: $ret");
  }

  return $ret;
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
sub cddb_down {
  close $socket or
    return (undef, "Could not terminate connection with ".CDDB_HOST." ".CDDB_PORT.": $!");
  return "Connection terminated";
}

# ---------------------------------------------------------------------------
# --> cddb hello username hostname clientname version
# <-- code hello and welcome username@hostname running clientname version
#
# codes:
#
#      200 ... Handshake successful
#      431 ... Handshake not successful, closing connection
#      402 ... Already shook hands
#      500 ... Syntax error
# ---------------------------------------------------------------------------
sub hello {
  cddb_up();

  my $username = $ENV{USER} || $ENV{LOGNAME} || "ripcduser";
  my $hostname = $ENV{HOST} || $ENV{HOSTNAME} || "localhost";

  # Send cddb query.
  print $socket "cddb hello $username $hostname ripcd 0.1\n";
  # Read cddb response.
  my $ret = <$socket>;

  my ($code) = $ret =~ /^(\d\d\d)\s+/;

  if ($code != 200) {
    cddb_down();
    return (undef, "$ret");
  }

  return $ret;
}

# ---------------------------------------------------------------------------
# --> cddb lscat
# <-- code Okay category list follows
#
# codes:
#
#      210	Okay category list follows (until terminating marker)
# ---------------------------------------------------------------------------
sub lscat {
  # Send cddb query.
  print $socket "cddb lscat\n";
  # Read cddb response.
  my $ret = <$socket>;

  my ($code) = $ret =~ /^(\d\d\d)\s+/;

  unless ($code == 210) {
    return (undef, $ret);
  }

  my @lar;

  while (<$socket>) {
    last if (/^\./);
    push(@lar, $_);
  }

  return join('', @lar);
}

# ---------------------------------------------------------------------------
# Provided that a handshake has already taken place (via &hello) and the cddb
# query succeeds, &query returns an array reference that can be used on the
# calling side as follows:
#
#      my ($ok, $err) = tbx::CDDB::query();
#
#      if (! $ok) {
#        die $err;
#      } else {
#        print "query ok: " . join('', @{$ok});
#      }
#
# --> cddb query discid ntrks off1 off2 ... nsecs
# <-- code (categ discid dtitle|close matches found)
#
# codes:
#
#      200	Found exact match
#      211	Found inexact matches, list follows (until terminating marker)
#      202	No match found
#      403	Database entry is corrupt
#      409	No handshake
# ---------------------------------------------------------------------------
sub query {
  unless (defined($id)) {
    id();
  }

  my $offsets = "";
  for (my $i = 0; $i < $#toc; $i++) {
      $offsets .= "$toc[$i]->{o}";
      if ($i < $#toc - 1) {
        $offsets .= " ";
      }
  }

  my $seconds = int(($toc[$#toc]->{o})/75);

  # Send cddb query.
  print STDERR "cddb query $id $#toc $offsets $seconds\n";
  print $socket "cddb query $id $#toc $offsets $seconds\n";
  # Read cddb response.
  my $ret = <$socket>;

  my ($code) = $ret =~ /^(\d\d\d)\s+/;

  unless ($code == 200 || $code == 211) {
    return (undef, $ret);
  }

  my @lar;

  if ($code == 200) {
    push(@lar, $ret);
    return \@lar;
  } elsif ($code == 211) {
    while(<$socket>) {
      last if (/^\./);
      push(@lar, $_);
    }

    return \@lar;
  } else {
    return (undef, $ret);
  }
}

# ---------------------------------------------------------------------------
# Find the cache file for the current working compact disc or for an arbitrary
# CD (if $cat and $eid are specified). Write $xmcd to the cache file if $write
# is true, otherwise set $xmcd equal to the file contents. Return $xmcd or ''.
# If cache returns '' when $write is true, there was probably a problem
# writing the file.
# ---------------------------------------------------------------------------
sub cache {
  my ($write, $cat, $eid) = @_;
  my $cache = "$ENV{HOME}/.cddb";

  $cat ||= $categ;
  $eid ||= $discid;

  unless (-e $cache) {
    system("mkdir -p $cache");
    unless (-e $cache) {
      return '';
    }
  }

  unless (-d $cache) {
    return '';
  }

  $cache = "$cache/$cat";
  unless (-e "$cache") {
    system("mkdir -p $cache");
    unless (-e $cache) {
      return '';
    }
  }

  unless (-d $cache) {
    return '';
  }

  $cache = "$cache/$eid";

  if (-e "$cache") { 
    unless ($write) {
      $xmcd = '';
      open(CACHE, $cache) or return ''; # [!] Should we return an error here?
      while (<CACHE>) {
        $xmcd .= $_;
      }
      close(CACHE);
      return $xmcd;
    }
  }

  if ($write && $xmcd ne '' && (! -e $cache || -w $cache)) {
    open(CACHE, ">$cache") or return ''; # [!] Should we return an error here?
    print CACHE $xmcd;
    close(CACHE);
    return $xmcd;
  }

  return '';
}

# ---------------------------------------------------------------------------
# --> cddb read categ discid
# <-- code (categ discid|categ discid No such CD entry in database)
#
# codes:
#
#      210	OK, CDDB database entry follows (until terminating marker)
#      401	Specified CDDB entry not found.
#      402	Server error.
#      403	Database entry is corrupt.
#      409	No handshake.
# ---------------------------------------------------------------------------
sub read {
  ($categ, $discid) = @_;

  my $str = cache(0, $categ, $discid);

  unless ($str) {
    # Send cddb query.
    print $socket "cddb read $categ $discid\n";
    # Read cddb response.
    my $ret = <$socket>;

    my ($code) = $ret =~ /^(\d\d\d)\s+/;

    unless ($code == 210) {
      return (undef, $ret);
    }

    while(<$socket>) {
      last if (/^\./);
      $str .= $_;
    }
    $xmcd = $str;
    cache(1);
  }

  return $xmcd;
}

# ---------------------------------------------------------------------------
# Parse the xcmd CD database file formatted string $xmcd and create a new
# album object for all the data.
# ---------------------------------------------------------------------------
sub album {
  my @lar = split('\n', $xmcd);

  my $DTITLE = '';
  my @TTITLE;
  $extd = '';
  @extt = ();

  for (@lar) {
    my $rec_sep = $/;
    $/ = "\r";
    chomp();
    $/ = $rec_sep;

    if (/^# Revision: (\d+)/) {
      $revision = $1;
    }

    if (/^#/) {
      next;
    }

    if (/^DISCID=/) {
      next;
    }

    if (/^DYEAR=(.*)$/) {
      $dyear = $1;
      next;
    }

    if (/^DGENRE=(.*)$/) {
      $dgenre = $1;
      next;
    }
    
    if (/^DTITLE=(.*)$/) {
      $DTITLE .= $1;
      next;
    }

    if (/^TTITLE(\d+)=(.*)/) {
      if ($TTITLE[$1]) {
        $TTITLE[$1] .= $2;
      } else {
        $TTITLE[$1] = $2;
      }
      next;
    }

    if (/^EXTD=(.*)$/) {
      $extd .= $1;
      next;
    }

    if (/^EXTT(\d+)=(.*)$/) {
      if ($extt[$1]) {
        $extt[$1] .= $2;
      } else {
        $extt[$1] = $2;
      }
    }

    if (/^(PLAYORDER=.*)$/) {
      $playorder .= $1;
      next;
    }
  }

  my ($artist, $title) = split(" / ", $DTITLE);

  $album = new tbx::Album($id, $artist, $title);
  
  $album->length(int(($toc[$#toc]->{o})/75));
  $album->year($dyear);
  $album->genre($dgenre);
  
  my $i = 1;

  open(TMP, ">/tmp/album");
  for (@TTITLE) {
    # [!] How are data tracks identified?
    # next if $info_tracks->[$i - 1]->is_data(); 

    my $o1 = $toc[$i]->{o};
    my $o2 = $toc[$i-1]->{o};

    my $t1 = ($o1 - ($o1 % CD_FRAMES)) / CD_FRAMES;
    my $t2 = ($o2 - ($o2 % CD_FRAMES)) / CD_FRAMES;
    
    my $s = ($t1 - $t2) % 60;
    my $m = int(($t1 - $t2) / 60);

    my ($artist, $title) = split(" / ", $_);

    if (! $title) {
      $title = $artist || '';
      $artist = '';
    }

    $artist ||= $album->artist();

    my $track = $album->add_track($artist, $title);

    $track->album($album->title());
    $track->freedb_id($album->freedb_id());
    $track->comment($album->freedb_id() . "::");
    $track->number($i);
    $track->position($i);
    $track->minutes($m);
    $track->seconds($s);
    $track->rip(1);
    $track->ripped(0);
    $track->year($album->year()) if $album->year();
    # What good is the genre is the id3 list doesn't include downtempo? Maybe
    # v2 will adress the shortcoming.
    # $track->year($album->genre());

    $i++;
  }
  close(TMP);
  return $album;
}

# ---------------------------------------------------------------------------
# Generate a string in the xmcd CD database file format
# ---------------------------------------------------------------------------
sub xmcd {
  my $str = '';
  my $col = 79;

  # # xmcd
  # #
  $str .= "# xmcd\n";
  $str .= "# \n";
  
  # # Track frame offsets:
  # #       150
  # #       ...
  # #
  $str .= "# Track frame offsets:\n";
  for my $n (0 .. scalar(@toc)-2) {
    $str .= "#       $toc[$n]->{o}\n";
  }
  $str .= "# \n";

  # # Disk length: N seconds
  # #
  $str .= "# Disc length: ". $album->length() ." seconds\n";
  $str .= "# \n";
  
  # # Revision: N
  # # Submitted via: rip-utils 0.1
  # #
  $str .= "# Revision: $revision\n";
  $str .= "# Submitted via: rip-utils 0.1\n";
  $str .= "# \n";
  
  # DISCID=0c19de14
  $str .= "DISCID=" . $album->freedb_id() . "\n";
  
  # DTITLE=Jimi Hendrix / The Ultimate Experience
  my $val = $album->artist() . " / " . $album->title();
  $str .= wrap_xmcd_str($val, "DTITLE=") . "\n";
  
  # DYEAR=1993
  if ($album->year()) {
    $str .= "DYEAR=". $album->year() . "\n";
  }

  # DGENRE=Classic Rock
  if ($album->genre()) {
    $str .= "DGENRE=". $album->genre() . "\n";
  }

  # TTITLE0=All Along the Watch Tower
  # ...
  my $pos = 0;
  for my $n (1 .. $album->number_of_tracks()) {
    if ($album->{$n}->artist() eq $album->artist()) {
      $val = $album->{$n}->title();
      $str .= wrap_xmcd_str($val, "TTITLE${pos}=") . "\n";
    } else {
      $val = $album->{$n}->artist() . " / " . $album->{$n}->title();
      $str .= wrap_xmcd_str($val, "TTITLE${pos}=") . "\n";
    }
    $pos++;
  }

  # EXTD=
  $str .= wrap_xmcd_str($extd, "EXTD=") . "\n";

  # EXTT0=
  # ...
  for my $n (0 .. $album->number_of_tracks()-1) {
    $val = $extt[$n] ? $extt[$n] : '';
    $str .= wrap_xmcd_str($val, "EXTT${n}=") . "\n";

  }

  # PLAYORDER=
  $str .= "PLAYORDER=\n";
  
  $xmcd = $str;
  cache(1);
}

# ---------------------------------------------------------------------------
# The simlest wrap routine possible. Add a line feed every 80th character.
# ---------------------------------------------------------------------------
sub wrap_xmcd_str {
  my ($str, $sep) = @_;
  my $col = 80 - length($sep);
  my $wstr = '';
  while (length($str) > $col) {
    $wstr .= ($wstr ? "\n" : '') . $sep . substr($str, 0, $col, '');
  }
  $wstr .= ($wstr ? "\n" : '') . $sep . $str;
  return $wstr;
}


# ---------------------------------------------------------------------------
# --> quit
# <-- code hostname Closing connection.  Goodbye.
# ---------------------------------------------------------------------------
sub goodbye {
  if ($socket) {
    print $socket "quit\n";
    my $ret = <$socket>;
    cddb_down();
    return $ret;
  }
}

1;