#!/usr/bin/perl

# rip-utils
#
# Copyright (C) 2001-2002 Mikol Graves
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA.

# -----------------------------------------------------------------------------
package tbx::FreeDB;
# -----------------------------------------------------------------------------

use strict;

use tbx::Album;

use IO::Socket;
use vars qw( @ISA @EXPORT );

@ISA = qw( Exporter );
@EXPORT = qw( &id );

sub CD_MSF_OFFSET       () { 150 }     # From /usr/include/linux/cdrom.h
sub CD_FRAMES           () { 75 }      # From /usr/include/linux/cdrom.h

sub FREEDB_MODE         () { "cddb" }
sub FREEDB_HOST         () { "freedb.freedb.org" }
sub FREEDB_PORT         () { 888 }

# Globals
my @toc;            # CD table of contents
my $id;             # CD's calculated FreeDB ID
my $album;          # CD's tbx::Album object
my $xmcd = '';      # xmcd CD database file
my $categ = '';     # effective FreeDB category
my $discid = '';    # effective FreeDB ID
my $revision = 0;   # revision number - increment when submitting to FreeDB
my $dyear = 0;      # xmcd DYEAR field
my $dgenre = '';    # xmcd DGENRE field
my $extd = '';      # xmcd EXTD field
my @extt;           # xmcd EXTTN fields
my $playorder = ''; # xmcd PLAYORDER field

my $freedb_mode = FREEDB_MODE;
my $freedb_host = FREEDB_HOST;
my $freedb_port = FREEDB_PORT;

my $http_proxy;
my $proxy_host;
my $proxy_port;

if ($ENV{http_proxy}) {
  proxy($ENV{http_proxy});
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub mode {
  my $mode = shift();
  if ($mode =~ /^cddb|http$/i) {
    $freedb_mode = $mode;
  }
  return $freedb_mode;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub host {
  my $host = shift();
  if ($host ne '') {
    $freedb_host = $host;
  }
  return $freedb_host;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub port {
  my $port = shift();
  if ($port =~ /^\d+$/) {
    $freedb_port = $port;
  }
  return $freedb_port;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub proxy {
  my $proxy = shift();
  if ($proxy eq '') {
    $http_proxy = '';
    $proxy_host = '';
    $proxy_port = '';
  } if ($proxy =~ /^(http:\/\/|)(.+?):(\d+)\/?$/i) {
    $proxy_host = $2;
    $proxy_port = $3;
    $http_proxy = $proxy;
  }
  return $http_proxy;
}

# -----------------------------------------------------------------------------
# Generate a valid FreeDB disc ID and store it in the $id global.
# -----------------------------------------------------------------------------
sub id {
  if ($id = shift()) {
    return $id;
  }
  $id = undef;
  my ($ok, $err) = toc();
  
  unless (defined($ok)) {
    return (undef, $err);
  }

  my $tracks = scalar(@toc) - 1;
  my $n = 0;
  my $t = 0;

  for (my $i = 0; $i < $tracks; $i++) {
    $n = $n + sum(($toc[$i]->{m} * 60) + $toc[$i]->{s});
  }

  unless ($#toc > -1) { 
    return (undef, "Couldn't determine CD table of contents." .
            " Maybe there is no disc in the drive.");
  }

#   $t = int(($toc[$#toc]->{o} - $toc[0]->{o})/75);
  $t = (($toc[$tracks]->{m} * 60) + $toc[$tracks]->{s}) -
    (($toc[0]->{m} * 60) + $toc[0]->{s});

  $id = sprintf("%08x", (($n % 0xff) << 24 | $t << 8 | $tracks));

  return $id;
}

# -----------------------------------------------------------------------------
# m ... minutes
# s ... seconds
# f ... frames
# o ... offset
# -----------------------------------------------------------------------------
sub toc {
  @toc = ();
  my $cdparanoia = `which cdparanoia`;
  chomp($cdparanoia);

  if ($cdparanoia) {
    if (-x $cdparanoia) {
      open(CDPARANOIA, "$cdparanoia -Q 2>&1 |");

      while (<CDPARANOIA>) {
        if (/^\s*\d+\.\s*\d+ \[.*?\]\s*(\d+) \[.*$/) {
          my ($m, $s, $f, $o) = lba_to_msf($1);
          push (@toc, { "m" => $m, "s" => $s, "f"  => $f, "o"  => $o });
        }
        if (/^TOTAL\s*(\d+) \[.*$/) {
          my ($m, $s, $f, $o) = lba_to_msf($1);
          push (@toc, { "m" => $m, "s" => $s, "f"  => $f, "o"  => $o });
        }
      }

      close(CDPARANOIA);

      return 1;
    } else {
      return (undef, "You do not have permission to execute cdparanoia.");
    }
  } else {
    return (undef, "The cdparanoia executable is not in your path.");
  }
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub lba_to_msf {
  my $o = shift();
  my ($m, $s, $f, $t);

  $o += CD_MSF_OFFSET;
  $f = $o % CD_FRAMES;

  $t = ($o - $f) / CD_FRAMES;

  $s = $t % 60;
  $m = int($t / 60);

  return ($m, $s, $f, $o);
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub sum {
  my $n = shift();
  my $ret = 0;

  while ($n > 0) {
    $ret += ($n % 10);
    $n = int($n / 10);
  }

  return $ret;
}

# FreeDB Client Subroutines ---------------------------------------------------
# --> Client to Server
# <-- Server to Client
# -----------------------------------------------------------------------------

my $socket;

# -----------------------------------------------------------------------------
# categ_and_discid will try to determine the FreeDB category and disc ID given
# the response from a FreeDB server. It's intended to be used with the value(s)
# returned by query().
# -----------------------------------------------------------------------------
sub categ_and_discid {
  my ($str) = @_;
  my ($code, $categ, $discid) = split(/\s/, $str);
  if ($code !~ /^\d\d\d$/) {
    $discid = $categ;
    $categ = $code;
  }
  return ($categ, $discid);
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub freedb_up {
  return $socket if $socket;

  my $host = $freedb_host;
  my $port = $freedb_port;

  if ($freedb_mode =~ /^http$/i && $http_proxy) {
    if ($proxy_host) {
      $host = $proxy_host;
    }

    if ($proxy_port){
      $port = $proxy_port;
    }
  }

  $socket = new IO::Socket::INET(PeerAddr => $host,
                                 PeerPort => $port,
                                 Proto    => "tcp",
                                 Type     => SOCK_STREAM);
  unless($socket) {
    return (undef, "Could not establish connection with " .
            $host . " " . $port . ": $!");
  }

  if ($freedb_mode =~ /^cddb$/i) {
    my $ret = <$socket>;

    unless ($ret =~ /^2\d\d\s+/) {
      return (undef, "Connection refused: $ret");
    }
    return $ret;
  }

  return 1;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub freedb_down {
  my $host = $freedb_host;
  my $port = $freedb_port;

  if ($freedb_mode =~ /^http$/i && $http_proxy) {
    if ($proxy_host) {
      $host = $proxy_host;
    }

    if ($proxy_port){
      $port = $proxy_port;
    }
  }

  if (defined($socket)) {
    close $socket or
      return (undef, "Could not terminate connection with " .
              $host . " " . $port . ": $!");

    $socket = undef;
  }

  return "Connection terminated.";
}

sub command {
  my $cmd = shift();

  if ($freedb_mode =~ /^http$/i) {
    my $hello = hello();
    $cmd =~ s/ /+/g;
    $cmd = "/~cddb/cddb.cgi?cmd=$cmd&hello=$hello&proto=5";
    if($http_proxy) {
      $cmd = "http://$freedb_host:$freedb_port$cmd HTTP/1.0\n";
    }
    $cmd = "GET $cmd";

    my ($ok, $err) = freedb_up();
    unless ($ok) {
      return ($ok, $err);
    }
  }

  print $socket "$cmd\n";

  if($freedb_mode =~ /^http$/i && $http_proxy) {
    while (<$socket> =~ /^\S+/) {
      ;
    }
  }

  return 1;
}

sub cleanup {
  my $ret = shift();
  my $str = shift();
  my ($d1, $d2, $d3, $sub) = caller(1);

  if ($freedb_mode =~ /^http$/i) {
    my ($ok, $err) = freedb_down();
    unless ($ok) {
      return ($ok, $ret ? $err : "$str $err");
    }
  }
  return ($ret, $str);
}

# -----------------------------------------------------------------------------
# --> cddb hello username hostname clientname version
# <-- code hello and welcome username@hostname running clientname version
#
# codes:
#
#      200 ... Handshake successful
#      431 ... Handshake not successful, closing connection
#      402 ... Already shook hands
#      500 ... Syntax error
# -----------------------------------------------------------------------------
sub hello {
  if ($freedb_mode =~ /^cddb$/i) {
    my ($ok, $err) = freedb_up();
    unless (defined($ok)) {
      return (undef, $err);
    }
  }

  my $username = $ENV{USER} || $ENV{LOGNAME} || "ripcduser";
  my $hostname = $ENV{HOST} || $ENV{HOSTNAME} || "localhost";
  my $hello = "$username $hostname ripcd 0.1";

  if ($freedb_mode =~ /^cddb$/i) {
    # Send freedb query.
    print $socket "cddb hello $hello\n";
    # Read freedb response.
    my $ret = <$socket>;

    my ($code) = $ret =~ /^(\d\d\d)\s+/;

    if ($code != 200) {
      freedb_down();
      return (undef, "$ret");
    }

    print $socket "proto 5\n";

    $ret = <$socket>;

    ($code) = $ret =~ /^(\d\d\d)\s+/;

    if ($code != 201) {
      freedb_down();
      return (undef, "$ret");
    }

    return $ret;
  } else {
    $hello =~ s/ /+/g;
    return $hello;
  }
}

# -----------------------------------------------------------------------------
# --> cddb lscat
# <-- code Okay category list follows
#
# codes:
#
#      210	Okay category list follows (until terminating marker)
# -----------------------------------------------------------------------------
sub lscat {
  my ($cmd, $err) = command("cddb lscat");
  unless ($cmd) {
    return cleanup(undef, $err);
  }

  # Read freedb response.
  my $ret = <$socket>;

  my ($code) = $ret =~ /^(\d\d\d)\s+/;

  unless (defined($code) && $code == 210) {
    if (! defined($code)) {
      while (<$socket>) {
        $ret .= $_;
      }
    }
    return cleanup(undef, $ret);
  }

  my @lar;

  while (<$socket>) {
    last if (/^\./);
    chomp();
    push(@lar, $_);
  }

  return cleanup(\@lar);
}

# -----------------------------------------------------------------------------
# Provided that a handshake has already taken place (via &hello) and the freedb
# query succeeds, &query returns an array reference that can be used on the
# calling side as follows:
#
#      my ($ok, $err) = tbx::FreeDB::query();
#
#      if (! $ok) {
#        die $err;
#      } else {
#        print "query ok: " . join('', @{$ok});
#      }
#
# --> cddb query discid ntrks off1 off2 ... nsecs
# <-- code (categ discid dtitle|close matches found)
#
# codes:
#
#      200	Found exact match
#      211	Found inexact matches, list follows (until terminating marker)
#      202	No match found
#      403	Database entry is corrupt
#      409	No handshake
# -----------------------------------------------------------------------------
sub query {
  unless (defined($id)) {
    id();
  }

  my $offsets = "";
  for (my $i = 0; $i < $#toc; $i++) {
      $offsets .= "$toc[$i]->{o}";
      if ($i < $#toc - 1) {
        $offsets .= " ";
      }
  }

  my $seconds = int(($toc[$#toc]->{o})/75);

  my ($cmd, $err) = command("cddb query $id $#toc $offsets $seconds");
  unless ($cmd) {
    return cleanup(undef, $err);
  }

  # Read freedb response.
  my $ret = <$socket>;

  my ($code) = $ret =~ /^(\d\d\d)\s+/;

  unless (defined($code) && ($code == 200 || $code == 211 || $code == 210)) {
    if (! defined($code)) {
      while (<$socket>) {
        $ret .= $_;
      }
    }
    return cleanup(undef, $ret);
  }

  my @lar;

  if ($code == 200) {
    push(@lar, $ret);
    return cleanup(\@lar);
  } elsif ($code == 211 || $code == 210) {
    while(<$socket>) {
      last if (/^\./);
      push(@lar, $_);
    }

    return cleanup(\@lar);
  } else {
    return cleanup(undef, $ret);
  }
}

# -----------------------------------------------------------------------------
# Find the cache file for the current working compact disc or for an arbitrary
# CD (if $cat and $eid are specified). Write $xmcd to the cache file if $write
# is true, otherwise set $xmcd equal to the file contents. Return $xmcd or ''.
# If cache returns '' when $write is true, there was probably a problem
# writing the file.
# -----------------------------------------------------------------------------
sub cache {
  my ($write, $cat, $eid) = @_;
  my $cache = "$ENV{HOME}/.cddb";

  $cat ||= $categ;
  $eid ||= $discid;

  unless (-e $cache) {
    system("mkdir -p $cache");
    unless (-e $cache) {
      return '';
    }
  }

  unless (-d $cache) {
    return '';
  }

  $cache = "$cache/$cat";
  unless (-e "$cache") {
    system("mkdir -p $cache");
    unless (-e $cache) {
      return '';
    }
  }

  unless (-d $cache) {
    return '';
  }

  $cache = "$cache/$eid";

  if (-e "$cache") { 
    unless ($write) {
      $xmcd = '';
      open(CACHE, $cache) or return ''; # [!] Should we return an error here?
      while (<CACHE>) {
        $xmcd .= $_;
      }
      close(CACHE);
      return $xmcd;
    }
  }

  if ($write && $xmcd ne '' && (! -e $cache || -w $cache)) {
    open(CACHE, ">$cache") or return ''; # [!] Should we return an error here?
    print CACHE $xmcd;
    close(CACHE);
    return $xmcd;
  }

  return '';
}

# -----------------------------------------------------------------------------
# --> cddb read categ discid
# <-- code (categ discid|categ discid No such CD entry in database)
#
# codes:
#
#      210	OK, FreeDB database entry follows (until terminating marker)
#      401	Specified FreeDB entry not found.
#      402	Server error.
#      403	Database entry is corrupt.
#      409	No handshake.
# -----------------------------------------------------------------------------
sub read {
  ($categ, $discid, my $nocache, my $force_lookup) = @_;

  my $str = $force_lookup ? '' : cache(0, $categ, $discid);

  unless ($str) {
    unless ($socket) {
      if ($freedb_mode =~ /^cddb$/i) {
        return $xmcd = '';
      }
    }

    my ($cmd, $err) = command("cddb read $categ $discid");
    unless ($cmd) {
      return cleanup(undef, $err);
    }

    # Read freedb response.
    my $ret = <$socket>;

    my ($code) = $ret =~ /^(\d\d\d)\s+/;

    unless (defined($code) && $code == 210) {
      if (! defined($code)) {
        while (<$socket>) {
          $ret .= $_;
        }
      }
      return cleanup(undef, $ret);
    }

    while(<$socket>) {
      last if (/^\./);
      $str .= $_;
    }

    $xmcd = $str;
    unless ($nocache) {
      cache(1);
    }
    (my $ok, $err) = cleanup(1);
    unless ($ok) {
      return (undef, $err);
    }
  }

  return $xmcd;
}

# -----------------------------------------------------------------------------
# Parse the xcmd CD database file formatted string $xmcd and create a new
# album object for all the data.
# -----------------------------------------------------------------------------
sub album {
  my @lar = split('\n', $xmcd);

  my $DTITLE = '';
  my @TTITLE;
  $extd = '';
  @extt = ();

  my $in_toc = 0;

  for (@lar) {
    my $rec_sep = $/;
    $/ = "\r";
    chomp();
    $/ = $rec_sep;

    if (/^#\s+Track frame offsets:/) {
      unless (scalar(@toc)) {
        $in_toc = 1;
      }
      next;
    }

    if (/^#\s+Disc length: (\d+) seconds/) {
      my $offset = $1 * CD_FRAMES;
      if ($in_toc) {
        push (@toc, { "o" => $offset });
        $in_toc = 0;
      }
      next;
    }

    if ($in_toc) {
      if (/^#\s+(\d+)/) {
        push (@toc, { "o" => $1 });
        next;
      }
    }

    if (/^#\s+Revision: (\d+)/) {
      $revision = $1;
    }

    if (/^#/) {
      next;
    }

    if (/^DISCID=/) {
      next;
    }

    if (/^DYEAR=(.*)$/) {
      $dyear = $1;
      next;
    }

    if (/^DGENRE=(.*)$/) {
      $dgenre = $1;
      next;
    }
    
    if (/^DTITLE=(.*)$/) {
      $DTITLE .= $1;
      next;
    }

    if (/^TTITLE(\d+)=(.*)/) {
      if ($TTITLE[$1]) {
        $TTITLE[$1] .= $2;
      } else {
        $TTITLE[$1] = $2;
      }
      next;
    }

    if (/^EXTD=(.*)$/) {
      $extd .= $1;
      next;
    }

    if (/^EXTT(\d+)=(.*)$/) {
      if ($extt[$1]) {
        $extt[$1] .= $2;
      } else {
        $extt[$1] = $2;
      }
    }

    if (/^(PLAYORDER=.*)$/) {
      $playorder .= $1;
      next;
    }
  }

  my ($artist, $title);

  if (scalar(@lar)) {
    ($artist, $title) = split(" / ", $DTITLE);
  } else {
    ($artist, $title) = ('Unknown Artist', 'Unknown Title');
  }

  $album = new tbx::Album($id, $artist, $title, $categ, $discid);
  
  $album->length(int(($toc[$#toc]->{o})/75));
  $album->year($dyear);
  $album->genre($dgenre);
  
  unless (scalar(@TTITLE)) {
    for (my $i = 0; $i < $#toc; $i++) {
      push(@TTITLE, "Unknown Artist / Unknown Title " . ($i + 1));
    }
  }

  my $i = 1;
  for (@TTITLE) {
    # [!] How are data tracks identified?
    # next if $info_tracks->[$i - 1]->is_data(); 

    my $o1 = $toc[$i]->{o};
    my $o2 = $toc[$i-1]->{o};

    my $t1 = ($o1 - ($o1 % CD_FRAMES)) / CD_FRAMES;
    my $t2 = ($o2 - ($o2 % CD_FRAMES)) / CD_FRAMES;
    
    my $s = ($t1 - $t2) % 60;
    my $m = int(($t1 - $t2) / 60);

    my ($artist, $title) = split(" / ", $_);

    if (! $title) {
      $title = $artist || '';
      $artist = '';
    }

    $artist ||= $album->artist();

    my $track = $album->add_track($artist, $title);

    $track->album($album->title());
    $track->freedb_id($album->freedb_id());
    $track->comment($album->freedb_id() . "::" . $track->number());
    $track->number($i);
    $track->position($i);
    $track->minutes($m);
    $track->seconds($s);
    $track->rip(1);
    $track->ripped(0);
    $track->year($album->year()) if $album->year();
    # What good is the genre if the id3 list doesn't include downtempo? Maybe
    # v2 will adress the shortcoming.
    # $track->year($album->genre());

    $i++;
  }
  return $album;
}

# -----------------------------------------------------------------------------
# Based on the current album information, write the xmcd CD database file and
# set $xmcd equal to its contents.
# -----------------------------------------------------------------------------
sub xmcd {
  my $str = shift();
  $str ||= '';
  my $col = 79;

  if ($str eq '') {
    # # xmcd
    # #
    $str .= "# xmcd\n";
    $str .= "# \n";

    # # Track frame offsets:
    # #       150
    # #       ...
    # #
    $str .= "# Track frame offsets:\n";
    for my $n (0 .. scalar(@toc)-2) {
      $str .= "#       $toc[$n]->{o}\n";
    }
    $str .= "# \n";

    # # Disk length: N seconds
    # #
    $str .= "# Disc length: ". $album->length() ." seconds\n";
    $str .= "# \n";

    # # Revision: N
    # # Submitted via: rip-utils 0.1 ru experienced?
    # #
    $str .= "# Revision: $revision\n";
    $str .= "# Submitted via: rip-utils v0.2 ru experienced?\n";
    $str .= "# \n";

    # DISCID=0c19de14
    $str .= "DISCID=" . $album->freedb_id() . "\n";

    # DTITLE=Jimi Hendrix / The Ultimate Experience
    my $val = $album->artist() . " / " . $album->title();
    $str .= wrap_xmcd_str($val, "DTITLE=") . "\n";

    # DYEAR=1993
    if ($album->year()) {
      $str .= "DYEAR=". $album->year() . "\n";
    }

    # DGENRE=Classic Rock
    if ($album->genre()) {
      $str .= "DGENRE=". $album->genre() . "\n";
    }

    # TTITLE0=All Along the Watch Tower
    # ...
    my $pos = 0;
    for my $n (1 .. $album->number_of_tracks()) {
      if ($album->{$n}->artist() eq $album->artist()) {
        $val = $album->{$n}->title();
        $str .= wrap_xmcd_str($val, "TTITLE${pos}=") . "\n";
      } else {
        $val = $album->{$n}->artist() . " / " . $album->{$n}->title();
        $str .= wrap_xmcd_str($val, "TTITLE${pos}=") . "\n";
      }
      $pos++;
    }

    # EXTD=
    $str .= wrap_xmcd_str($extd, "EXTD=") . "\n";

    # EXTT0=
    # ...
    for my $n (0 .. $album->number_of_tracks()-1) {
      $val = $extt[$n] ? $extt[$n] : '';
      $str .= wrap_xmcd_str($val, "EXTT${n}=") . "\n";

    }

    # PLAYORDER=
    $str .= "PLAYORDER=\n";
  }

  $xmcd = $str;
  cache(1);
}

# -----------------------------------------------------------------------------
# The simplest wrap routine possible. Place a line feed every 80th character.
# -----------------------------------------------------------------------------
sub wrap_xmcd_str {
  my ($str, $sep) = @_;
  my $col = 80 - length($sep) - 1;
  my $wstr = '';
  while (length($str) > $col) {
    $wstr .= ($wstr ? "\n" : '') . $sep . substr($str, 0, $col, '');
  }
  $wstr .= ($wstr ? "\n" : '') . $sep . $str;
  return $wstr;
}


sub submit {
  my $email = shift();

  my $host = $freedb_host;
  my $port = $freedb_mode =~ /^http$/i ? $freedb_port : 80;

  if ($freedb_mode =~ /^http$/i && $http_proxy) {
    if ($proxy_host) {
      $host = $proxy_host;
    }

    if ($proxy_port){
      $port = $proxy_port;
    }
  }

  my $submit_socket = new IO::Socket::INET(PeerAddr => $host,
                                           PeerPort => $port,
                                           Proto    => "tcp",
                                           Type     => SOCK_STREAM);
  unless($submit_socket) {
    return (undef, "Could not establish connection with " .
            $host . " " . $port . ": $!");
  }

  $revision++;
  my $submit_xmcd = xmcd();
  my $submit_length = length($submit_xmcd);

  print $submit_socket "POST /~cddb/submit.cgi HTTP/1.0\n";
  print $submit_socket "Category: $categ\n";
  print $submit_socket "Discid: $discid\n";
  print $submit_socket "User-Email: $email\n";
  #   print $submit_socket "Submit-Mode: test\n";
  print $submit_socket "Submit-Mode: submit\n";
  print $submit_socket "X-Cddbd-Note: Submitted for $email by rip-utils. ru experienced?\n";
  print $submit_socket "Content-Length: $submit_length\n\n";

  print $submit_socket "$xmcd";

  print $submit_socket "\n\n";

  my $ret = '';

  while (<$submit_socket>) {
    $ret .= $_;
  }

  if (defined($submit_socket)) {
    close $submit_socket or
      return (undef, "Could not terminate connection with " .
              $host . " " . $port . ": $!");

    $submit_socket = undef;
  }

  return $ret;
}

# -----------------------------------------------------------------------------
# --> quit
# <-- code hostname Closing connection.  Goodbye.
# -----------------------------------------------------------------------------
sub goodbye {
  if ($socket && $freedb_mode =~ /^cddb$/i) {
    print $socket "quit\n";
    my $ret = <$socket>;
    freedb_down();
    return $ret;
  }
  return 1;
}

1;