#!/usr/bin/perl

# rip-utils
#
# Copyright (C) 2001-2002 Mikol Graves
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA.

# ---------------------------------------------------------------------------
package tbx::Kit;
# ---------------------------------------------------------------------------

use strict;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);
@EXPORT = qw(new);

# ~/.thinbox/rip-utils/rip-utilsrc usage
my %rc_comments = (
  CDROMDevice                  => <<END_MSG,

Use CDROMDevice to select which CD-ROM device to use.

END_MSG

  CategoriesDirectoryName      => <<END_MSG,

The name of the parent directory for tracks categorized by lnmp3. The
categories directory will be a child of the TargetDirectory.

END_MSG

  DefaultM3UFilename           => <<END_MSG,

The name of the playlist created my mkm3u if one is not specified on the
command line. The default playlist will be a child of the TargetDirectory.

END_MSG

  EmailAddress                 => <<END_MSG,

This field is required in order to submit corrections or additions to your
local FreeDB server. If a submission is rejected, the rejection notice will be
delivered to the address you specify here.

END_MSG

  EncCommand                   => <<END_MSG,

EncCommand is the command-line, including options, for LAME. Personal and
commercial use of compiled versions of LAME (or any other executible that can
encode mp3 files) requires a patent license in some countries.

END_MSG

  EnqCommand                   => <<END_MSG,

The default enqueue command used by rip-utils such as mkm3u when they want to
start, or add tracks to, a media player.
  
END_MSG

  FreeDBHost                   => <<END_MSG,

Contact the FreeDB server specified by FreeDBHost.

END_MSG

  FreeDBMode                   => <<END_MSG,

Communicate with the FreeDB server using the protocol specified by FreeDBMode.

END_MSG

  FreeDBPort                   => <<END_MSG,

Talk to the FreeDB server on the port specified by FreeDBPort. If FreeDBMode
is "http", FreeDBPort should usually be "80".

END_MSG

  M3UFilename                  => <<END_MSG,
  
M3UFilename is the template for the name of the playlist file for the current
album. Any slashes ('/') will be interpreted as demarcating directories. The
'.m3u' extension will be appended automatically.

END_MSG

  MP3Filename                  => <<END_MSG,

MP3Linkname is the template for the symbolic link name used by lnmp3 when
categorizing tracks. Any slashes ('/') will be interpreted as demarcating
directories. The '.mp3' extension will be appended automatically.

END_MSG

  MP3Linkname                  => <<END_MSG,
  
MP3Linkname is the template for the name of the encoded mp3 file for the
current track. Any slashes ('/') will be interpreted as demarcating
directories. The '.mp3' extension will be appended automatically.

END_MSG
  
  MaxEncCount                  => <<END_MSG,

Use MaxEncCount to limit the number of encoder processes. If this option is
left unspecified, its value will be equal to the number of processors as
determined by reading /proc/cpuinfo.

END_MSG

  MultiArtistMP3Filename       => <<END_MSG,
  
MultiArtistMP3Filename is used instead of MP3Filename when the album artist
and the track artist are different.

END_MSG

  ReplaceSpacesWithUnderscores => <<END_MSG,

If ReplaceSpacesWithUnderscores is 'true', spaces in expanded '%' switches
will be replaced with underscores.

END_MSG

  RipCommand                   => <<END_MSG,

RipCommand is the command line, including options, for cdparanoia. Because it
detects and corrects a variety of errors, cdparanoia is very robust but slow.
If you know that your device is relatively error-free and you are willing to
sacrifice robustness for performance, you can disable cdparanoia's 'paranoid'
features with the -Y and -Z options.

END_MSG

  SafeCharacters               => <<END_MSG,

SafeCharacters is a Perl regular expression of characters that will -not- be
escaped in file names. All other characters will be URI-escaped (e.g., ':'
will become '%3A'). The default will generate very portable file names.",

END_MSG

  TargetDirectory              => <<END_MSG,

Set TargetDirectory to the desired final destination of the encoded files.

END_MSG

  TemporaryDirectory           => <<END_MSG,

Set TemporaryDirectory to the directory of your choice. Please ensure that the
directory includes sufficient free space.

END_MSG

  UseTemporaryDirectory        => <<END_MSG,

If UseTemporaryDirectory is 'true', write ripped and encoded files to a
temporary directory before moving them to their final destination.

END_MSG
);

sub new {
  my $this = shift;
  my $class = ref($this) || $this;
  my $self = {};
  bless $self, $class;
  if (my $ret = $self->initialize()) {
    return (undef, $ret);
  }
  return $self;
}

sub initialize {
  my $this = shift();
  $this->{rc} = {
    CDROMDevice                  => qq?/dev/cdrom?,
    CategoriesDirectoryName      => qq?@ Categories?,
    DefaultM3UFilename           => qq?@ Playlist.m3u?,
    DebugLevel                   => qq?0?,
    EmailAddress                 => '',
    EncCommand                   => qq?lame --r3mix --nohist --id3v2-only --tl "%T" --ta "%a" --tt "%t" --tc "%i::%n::%c/%e" "%F" "%f"?,
    EnqCommand                   => qq?xmms?,
    FreeDBHost                   => qq?freedb.freedb.org?,
    FreeDBMode                   => qq?cddb?,
    FreeDBPort                   => qq?8880?,
    M3UFilename                  => qq?%A - %T?,
    MP3Filename                  => qq?%A/%T/%N. %t?,
    MP3Linkname                  => qq?%a - %t?,
    MaxEncCount                  => count_cpus(),
    MultiArtistMP3Filename       => qq?%A/%T/%N. %a - %t?,
    ReplaceSpacesWithUnderscores => qq?false?,
    RipCommand                   => qq?cdparanoia -w -e %n "%F"?,
    SafeCharacters               => qq?\\ a-zA-Z0-9\\@\\-_\\.\\(\\)'?,
    TargetDirectory              => qq?$ENV{HOME}/mp3/?,
    TemporaryDirectory           => qq?/tmp/?,
    UseTemporaryDirectory        => qq?true?
  };

  $this->{RIP_DIR} = "";
  $this->{ENCODED_FILE} = "";

  my $e;
  if ($e = readin_rc_file($this)) {
    return $e;
  }

  if ($e = check_kit($this)) {
    return $e;
  }
}

# readin_rc_file --------------------------------------------------------------

# Ensure that .thinbox/thinice/riprc exists in the current user's home
# directory. If the riprc file has just been created, populate it. Read in the
# riprc file and override any defaults found in %rc.

# -----------------------------------------------------------------------------
sub readin_rc_file {
  my $this = shift();
  my $rc_dir = "$ENV{HOME}/.thinbox/rip-utils/";
  unless(-e "$rc_dir") {
    `mkdir -p \"$rc_dir\"`;
    my $e = $!;
    unless (-e "$rc_dir") {
      return ("Could not make directory: $e\n");
    }
  }

  my $rc_file = "${rc_dir}rip-utilsrc";
  unless(-e "$rc_file") {
    `touch $rc_file`;
    my $e = $!;
    unless (-e "$rc_file") {
      return ("Could not make file: $e\n");
    }
    writeout_rc_defaults($this, $rc_file);
  }
  writeout_rc_defaults($this, "$rc_file.defaults");

  # Read options, overriding defaults.
  open RC_FILE, "<$rc_file";
  while (<RC_FILE>) {
    $_ =~ s/^\s*(.*?)\s*$/$1/;    # Trim white space.
    unless ($_ !~ /^\#/) {        # Skip comment lines.
      next;
    }
    if (/=/) {
      s/(\#.*)$//;
      my ($a,$v) = split("=");
      $a =~ s/^\s*(.*?)\s*$/$1/;  # Trim white space.
      $v =~ s/^\s*(.*?)\s*$/$1/;  # Trim white space.
      $this->{rc}{$a} = $v;
    }
  }
  close RC_FILE;

  # Make sure everbody knows where files should be ripped and encoded.
  if ($this->{rc}{"UseTemporaryDirectory"} =~ /true/i) {
    $this->{RIP_DIR} = $this->{rc}{"TemporaryDirectory"};
  } else {
    $this->{RIP_DIR} = $this->{rc}{"TargetDirectory"};
  }
  $this->{ENCODED_FILE} = "$this->{ENCODED_FILE}.thinice-rip.encoded";
  #   $catalog = thinBox::thinIce::Catalog->new("$this->{rc}{TargetDirectory}catalog.xml");
  return;
}

# writeout_rc_defaults --------------------------------------------------------

# Use the defaults in %rc and the comments in %rc_comments to generate a
# default and informative rc file. Do not include anything for undocumented
# features.

# -----------------------------------------------------------------------------
sub writeout_rc_defaults {
  my ($this, $rc_file) = @_;
  open RC_FILE, ">$rc_file";

  # Prologue the rc file with entity documentation.
  print RC_FILE "\# % Switch Usage ------------------------------------------------------------$/";
  print RC_FILE "\#$/";
  print RC_FILE "\# Where used, '%' switches will be replaced as follows:$/";
  print RC_FILE "\#$/";
  print RC_FILE "\# %A ... album artist$/";
  print RC_FILE "\# %T ... album title$/";
  print RC_FILE "\# %a ... track artist$/";
  print RC_FILE "\# %t ... track title$/";
  print RC_FILE "\# %N ... padded track number$/";
  print RC_FILE "\# %n ... track number$/";
  print RC_FILE "\# %F ... .wav file name$/";
  print RC_FILE "\# %f ... .mp3 file name$/$/";

  for my $k (sort(keys(%{$this->{rc}}))) {
    unless ($rc_comments{$k}) {
      # If an option doesn't have a comment, then it is experimental and
      # should not be documented.
      next;
    }
    # Create a pretty comment title for the current pref.
    my $t = "# $k ";
    $t .= "-" x (78 - length($t));
    print RC_FILE "$t$/";
    # Include documentation for the current pref.
    my $comment = "# $rc_comments{$k}";
    my $crlf = $/;
    $comment =~ s/(\Q$crlf\E)/$1\# /g;
    print RC_FILE "$comment";
    # Print the pref default as part of its documentation.
    my $default = $this->{rc}{$k};
    print RC_FILE "Default: $default$/";
    # Include a commented line that actually sets the pref default.
    my $pref = "# $k";
    $pref .= " " x (50 - length($pref)); # Make sure values are aligned.
    print RC_FILE "#$/$pref= $default$/$/";
  }
  close RC_FILE;
}

# check_kit -------------------------------------------------------------------

# Don't continue blindly only to fail at some later point.

# -----------------------------------------------------------------------------
sub check_kit {
  my $this = shift();

  if (-e $this->{rc}{"CDROMDevice"}) {
    unless (-r $this->{rc}{"CDROMDevice"}) {
      return ("You are not able to read from the configured CD-ROM device (CDROMDevice = \"$this->{rc}{CDROMDevice}\").");
    }
  } else {
    return ("The configured CD-ROM device (CDROMDevice = \"$this->{rc}{CDROMDevice}\") does not exist.");
  }

#   my $cddbrc_file = "$ENV{HOME}/.cddbrc";
#   my $cdserverrc_ln = "$ENV{HOME}/.cdserverrc";
#   if (-e "$cddbrc_file") {
#     unless (-e "$cdserverrc_ln") {
#       `ln -s \"$cddbrc_file\" \"$cdserverrc_ln\"`;
#       my $err = $!;
#       unless (-e "$cdserverrc_ln") {
#         return ("Could not link $cdserverrc_ln to $cddbrc_file".($err?": $err":"").".");
#       }
#     }
#   } else {
#     open CDDBRC_FILE, ">$cddbrc_file";
#     print CDDBRC_FILE "SERVER=$this->{rc}{CDDBServer}\n";
#     print CDDBRC_FILE "PROXY=$this->{rc}{ProxyServer}\n" if $this->{rc}{ProxyServer} ne "";
#     close CDDBRC_FILE;
#     if (-e "$cddbrc_file") {
#       unless (-e "$cdserverrc_ln") {
#         `ln -s \"$cddbrc_file\" \"$cdserverrc_ln\"`;
#         my $err = $!;
#         unless (-e "$cdserverrc_ln") {
#         return ("Could not link $cdserverrc_ln to $cddbrc_file".($err?": $err":"").".");
#         }
#       }
#     } else {
#       my $err = $!;
#       return ("Could not create $cddbrc_file".($err?": $err":"").".");
#     }
#  }

  unless (($this->{rc}{"UseTemporaryDirectory"} =~ /^false$/i)) {
    if (-e $this->{rc}{"TemporaryDirectory"}) {
      unless (-d $this->{rc}{"TemporaryDirectory"}) {
        return ("The configured temporary directory (TemporaryDirectory = \"$this->{rc}{TemporaryDirectory}\") is not a directory.");
      }
      unless (-w $this->{rc}{"TemporaryDirectory"}) {
        return ("You are not able to write to the configured temporary directory (TemporaryDirectory = \"$this->{rc}{TemporaryDirectory}\").");
      }
    } else {
      `mkdir -p \"$this->{rc}{TemporaryDirectory}\"`;
      my $err = $!;
      unless (-e "$this->{rc}{TemporaryDirectory}") {
        return ("Could not create temporary directory (TemporaryDirectory = \"$this->{rc}{TemporaryDirectory}\")".($err?": $err":"").".");
      }
    }
  }

  if (-e $this->{rc}{"TargetDirectory"}) {
    unless (-d $this->{rc}{"TargetDirectory"}) {
      return ("The configured target directory (TargetDirectory = \"$this->{rc}{TargetDirectory}\") is not a directory.");
    }
    unless (-w $this->{rc}{"TargetDirectory"}) {
      return ("You are not able to write to the configured target directory (TargetDirectory = \"$this->{rc}{TargetDirectory}\").");
    }
  } else {
    `   mkdir -p \"$this->{rc}{TargetDirectory}\"`;
    my $err = $!;
    unless (-e "$this->{rc}{TargetDirectory}") {
      return ("Could not create target directory (TargetDirectory = \"$this->{rc}{TargetDirectory}\")".($err?": $err":"").".");
    }
  }

  my $rip = $this->{rc}{"RipCommand"};
  $rip =~ s/^(.*?)(-|%)$/$1/;
  my $rip_ret = `which $rip`;
  chomp($rip_ret);
  if ($rip_ret) {
    unless (-x $rip_ret) {
      return ("You do not have permission to execute the configured ripping application (RipCommand = \"$this->{rc}{RipCommand}\").");
    }
  } else {
    return ("The configured ripping application (RipCommand = \"$this->{rc}{RipCommand}\") does not exist, is not in your path or you do not have permission to execute it.");
  }

  my $enc = $this->{rc}{"EncCommand"};
  $enc =~ s/^(.*?)(-|%)$/$1/;
  my $enc_ret = `which $enc`;
  chomp($enc_ret);
  if ($enc_ret) {
    unless (-x $enc_ret) {
      return ("You do not have permission to execute the configured encoding application (EncCommand = \"$this->{rc}{EncCommand}\").");
    }
  } else {
    return ("The configured encoding application (EncCommand = \"$this->{rc}{EncCommand}\") does not exist, is not in your path or you do not have permission to execute it.");
  }

  my $id3v2_ret = `which id3v2`;
  chomp($id3v2_ret);
  if ($id3v2_ret) {
    unless (-x $id3v2_ret) {
      return ("You do not have permission to execute \"$id3v2_ret\").");
    }
  } else {
    return ("\"id3v2\" does not exist, is not in your path or you do not have permission to execute it.");
  }

  return;
}

# count_cpus ------------------------------------------------------------------

# Read /proc/cpuinfo looking for lines that begin with 'processor'. Return 1
# if we can't read we misread /proc/cpuinfo.

# -----------------------------------------------------------------------------
sub count_cpus {
  my $count = 0;
  open CPUINFO, "</proc/cpuinfo" or return 1;
  while (<CPUINFO>) {
    $count++ if (/^processor\s+\:\s\d+$/);
  }
  close CPUINFO;
  return $count or 1;
}

# -----------------------------------------------------------------------------
# Read/Write the current working compact disk file.
# -----------------------------------------------------------------------------
sub cwcd {
  my ($this, $id, $cat, $eid) = @_;

  my $rc_dir = "$ENV{HOME}/.thinbox/rip-utils/";
  unless(-e "$rc_dir") {
    `mkdir -p \"$rc_dir\"`;
    my $e = $!;
    unless (-e "$rc_dir") {
      return ("Could not make directory: $e\n");
    }
  }

  my $cwcd_file = "${rc_dir}cwcd";

  if ($id && $cat && $eid) {
    open(CWCD, ">$cwcd_file") or
      return (undef, "Could not write current working compact disk file: $!");
    print CWCD "$id,$cat,$eid\n";
    close(CWCD);
  } else {
    open(CWCD, "<$cwcd_file") or return ('', '', '');
    my $str;
    while (<CWCD>) {
      chomp();
      $str = $_;
      last;
    }
    close(CWCD);
    ($id, $cat, $eid) = split(',', $str);
  }
  return ($id, $cat, $eid);
}

1;