#!/usr/bin/perl

# rip-utils
#
# Copyright (C) 2001-2002 Mikol Graves
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 59 Temple
# Place - Suite 330, Boston, MA 02111-1307, USA.

# ---------------------------------------------------------------------------
package tbx::RipUtil;
# ---------------------------------------------------------------------------

use strict;
use Getopt::Long();

use tbx::Kit();
use tbx::FreeDB();

my $kit;
my $util;
my $album;

my $cid;
my $eid;
my $cat;

my %backups;

my %escapes;
for (0..255) {
  $escapes{chr($_)} = sprintf("%%%02X", $_);
}

my $err;
($kit, $err) = new tbx::Kit();
unless (defined($kit)) {
  print STDERR "tbx::RipUtil: $err$/";
  exit(1);
}

# -----------------------------------------------------------------------------
# $k ... rip-utilsrc option key
# $v ... rip-utilsrc option value
#
# Return the rip-utilsrc value for key $k. If $v is defined, set the value 
# before returning.
# -----------------------------------------------------------------------------
sub rc {
  my ($k, $v) = @_;
  $kit->{rc}{$k} = $v if defined $v;
  return $kit->{rc}{$k};
}

my %basement;   # The base state of an opt-arg hash
my %segment;    # A hash of an opt-arg group
my @segments;   # A list of opt-arg groups

# -----------------------------------------------------------------------------
#  HASHREF ... Option defaults - excluding a default for range
# @options ... A list of Getopt::Long option specifications
#
# Populate the global container for the current opt-arg group (command line
# interface segment) and the global list of all segments on the current
# command line. cli_segments passes any unknown words to cli_token, which adds
# valid range and argument tokens to the current segment hash.
# -----------------------------------------------------------------------------
sub cli_segments {
  %basement = %{shift()}; 
  my @options   = @_;

  # Reset the segment hash and list of segments before continuing.
  $basement{'<>'} = \&cli_token;
  %segment = %basement;
  @segments = ();

  Getopt::Long::Configure("gnu_getopt",
                          "no_ignore_case",
                          "no_auto_abbrev",
                          "pass_through");
  Getopt::Long::GetOptions (\%segment, @options, '<>');

  # We need at least one segment, even if it is the default. If the current
  # segment is not a default segment, make sure to include it in the list.
  if (scalar(@segments) < 1 || exists($segment{range})) {
    my %hash = %segment;
    push (@segments, \%hash);
  }

  return @segments;
}

# -----------------------------------------------------------------------------
# $word ... a command line token
#
# Process -<range> (.e., -1,4-7,10) options and arbitrary arguments. Create
# new segments using range options as boundaries.
# -----------------------------------------------------------------------------
sub cli_token {
  my $word = shift();
  my $modified = 0;

  for my $key (keys(%segment)) {
    if (! exists($basement{$key})) {
      $modified = 1;
      last;
    }

    # Hashes and arrays do not exist by default so if we've reached this point
    # we can probably compare string values safely.
    if ($segment{$key} ne $basement{$key}) {
      $modified = 1;
    }
  }

  # $word is a range token.
  if ($word =~ /^-(-?\d+)/ || $word =~ /^--all$/i) {
    my $range = range_token($word);
    if ($range ne $word) {
      if ($modified) {
        my %hash = %segment;
        push(@segments, \%hash);
        %segment = %basement;
        $segment{range} = $range;
      } else {
        $segment{range} = $range;
      }
    } else {
      print STDERR "Invalid range: $range$/";
      exit(1);
    }
  }
  # $word is an invalid option.
  elsif ($word =~ /^-\D+/) {
    print STDERR "Invalid option: $word$/";
    exit(1);
  } 
  # $word is an argument.
  else {
    push(@{$segment{args}}, $word);
  }
}

# -----------------------------------------------------------------------------
# $word ... A token that may or may not be a valid range
#
# Return the range stripped of the option starter '-' if $word is a valid
# range option. Return $word otherwise.
# -----------------------------------------------------------------------------
sub range_token {
  my $word = shift();
  my $range = '';

  if ($word =~ /^--(all)$/i) {
    return $1;
  }

  $word =~ s/^--?//;
  my $str = $word;

  while ($str ne '') {
    if ($str =~ /^(\d+((-|,)\d+)*)/) {
      $str =~ s/(\d+((-|,)\d+)*)//;
      $range .= $1;
    } else {
      last;
    }
  }

  return $str ? $word : $range;
}

# -----------------------------------------------------------------------------
# $interactive ... Boolean indicating whether to consult the cache or not
#        $xmcd ... String specifying the location of an xmcd CD database file
#
# &freedb generates an informational object $album for the current disc.
# Unless directed to do otherwise via $interactive or $xmcd, &freedb will do
# its best to avoid contacting a freedb server unecessarily.
# -----------------------------------------------------------------------------
sub freedb {
  my $interactive = shift();
  my $xmcd = shift();
  my $force_lookup = shift();
  my $use_cache = 0;

  my $eject = `which eject`;
  chomp($eject);
  if ($eject) {
    if (-x $eject) {
      `$eject -t $kit->{rc}{CDROMDevice}`;
    }
  }

  my $mode = tbx::FreeDB::mode(rc('FreeDBMode'));
  unless ($mode eq rc('FreeDBMode')) {
    return (undef, "Could not set freedb mode.");
  }

  my $host = tbx::FreeDB::host(rc('FreeDBHost'));
  unless ($host eq rc('FreeDBHost')) {
    return (undef, "Could not set freedb host.");
  }

  my $port = tbx::FreeDB::port(rc('FreeDBPort'));
  unless ($port eq rc('FreeDBPort')) {
    return (undef, "Could not set freedb port.");
  }

  # Calculated ID
  my ($ok, $err) = tbx::FreeDB::id();
  unless ($ok) {
    if (! $xmcd) {
      return (undef, $err);
    } else {
      ($cat, $eid) = split('/', $xmcd);
      $ok = tbx::FreeDB::id($eid);
      if ($force_lookup) {
        tbx::FreeDB::read($cat, $eid, 1);
        tbx::FreeDB::album();
      }
    }
  }
  $cid = $ok;

  # Grab info about the last known CD from the cwcd file.
  my ($cwid, $cwcat, $cweid) = $kit->cwcd();

  # If the calculated ID and the current working ID match, generate an xmcd
  # cache file name using the current working category and effective ID.
  if (! $force_lookup && $cwid eq $cid && ! $interactive && ! $xmcd) {
    $xmcd = "$cwcat/$cweid";
    $use_cache = 1;
  }

  # Try to get the album information from a cache file without contacting a
  # freedb server.
  if (! $force_lookup && $xmcd ne '') {
    ($cat, $eid) = split('/', $xmcd);

    my $cached = tbx::FreeDB::read($cat, $eid) ? 1 : 0;
    
    # If we don't expect a cache file because the calculated ID and the
    # current working ID do not match or we found a cache file the user
    # specified, generate an album object using the cache file and do not
    # contact a freedb server.
    if (! $use_cache || $cached) {
      $album = tbx::FreeDB::album();
      # Update the cwcd file in case $xmcd was specified explicitly.
      $kit->cwcd($cid, $cat, $eid);
      return 1;
    }
  }

  # If we reach this point, there is no cache file or $interactive is true so
  # we need to contact a freedb server.
  ($ok, $err) = tbx::FreeDB::hello();

  if (! $ok) {
    return (undef, "$err. Perhaps you are not connected to the Internet.");
  }

  ($ok, $err) = tbx::FreeDB::query();

  # The calculated ID does not match any freedb entries.
  if (! $ok) {
    if ($err =~ /^202/) {
      my ($categs, $lscaterr) = tbx::FreeDB::lscat();
      if (! $categs) {
        $categs = $lscaterr;
      } else {
        for (@{$categs}) {
          $_ = "     $_";
        }
        $categs = join("$/", @{$categs});
      }

      $err .= <<END_MSG; 
Perhaps you should use the -x option to create a local entry using one of the 
following categories:

$categs

END_MSG

      $err .= "For example: -x misc/$cid$/";
    }
    tbx::FreeDB::goodbye();
    return (undef, "query error: $err");
  }

  # There's at least one matching freedb entry for the current disc.
  my @lar = @{$ok};

  my $default = 0;
  my $cache_hits = 0;

  my $question = "There are multiple database entries matching disc id $cid:\n";
  for my $n (0 .. $#lar) {
    my $cached = tbx::FreeDB::cache(0, tbx::FreeDB::categ_and_discid($lar[$n]));
    $question .= "\n  " . ($cached ? "cached" : "      ") . "  $n  $lar[$n]";
    if ($cached) {
      $default = $n;
      $cache_hits++;
    }
  }
  $question .= "\nPlease select one.";

  my $answer = scalar(@lar) > 1 
    ? ($cache_hits != 1 || $interactive ? prompt($question, $default) : $default)
    : $default;

  ($cat, $eid) = tbx::FreeDB::categ_and_discid($lar[$answer]);

  # As it stands now, $force_lookup is only used when we want to grab the
  # remote album info for a diff. In this case, we don't want to overwrite the
  # local cache file either: read(category, discid, nocache, force lookup).
  ($ok, $err) = tbx::FreeDB::read($cat, $eid, $force_lookup, $force_lookup);

  if (! $ok) {
    tbx::FreeDB::goodbye();
    return (undef, "read error: $err");
  }

  $album = tbx::FreeDB::album();

  ($ok, $err) = tbx::FreeDB::goodbye();

  if (! $ok) {
    return (undef, "goodbye error: $err");
  }

  # Update the cwcd file for the current working disk.
  ($ok, $err) = 
    $kit->cwcd($cid, $cat, $eid);

  if (! $ok) {
    return (undef, "cwcd error: $err");
  }

  return 1;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub album {
  my ($a) = @_;
  $album = $a if defined $a;
  return $album;
}

# -----------------------------------------------------------------------------
# Write out the xmcd CD database file for the current album
# -----------------------------------------------------------------------------
sub xmcd {
  tbx::FreeDB::xmcd();
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub backup {
  my $backup = "$ENV{HOME}/.thinbox/rip-utils/chcd-backup.$cat.$eid";
  if (! $backups{$backup}) {
    my $ok = open(BACKUP, ">$backup");
    unless ($ok) { 
      print STDERR "tbx::RipUtil: Unable to backup to \"$backup\": $!$/";
      exit(1);
    }
    print BACKUP tbx::FreeDB::cache();
    close(BACKUP);
    $backups{$backup} = 1;
  }
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub restore {
  my $backup = "$ENV{HOME}/.thinbox/rip-utils/chcd-backup.$cat.$eid";
  if (-e $backup) {
    my $restore = '';
    my $ok = open(BACKUP, "<$backup");
    unless ($ok) { 
      print STDERR "tbx::RipUtil: Unable restore from \"$backup\": $!$/";
      exit(1);
    }
    while (<BACKUP>) {
      $restore .= $_;
    }
    close(BACKUP);

    tbx::FreeDB::xmcd($restore);
    `rm \"$backup\"`;
  }
}

sub submit {
  return tbx::FreeDB::submit(shift());
}

# -----------------------------------------------------------------------------
# Use standard URI encoding to make file names more portable.
# -----------------------------------------------------------------------------
sub escape_unsafe_chars {
  my ($str) = @_;
  if (tbx::RipUtil::rc("ReplaceSpacesWithUnderscores") =~ /true/i) {
    $str =~ s/\s/_/g;
  }
  my $safe_chars = tbx::RipUtil::rc("SafeCharacters");
  $str =~ s/([^$safe_chars])/$escapes{$1}/g;
  return $str;
}

sub trever_xmcd {
  my $al = shift();
  my $alcat = $al->freedb_cat();
  my $aleid = $al->freedb_eid();
  my $cache = "$ENV{HOME}/.cddb";
  my $trever = "$ENV{HOME}/.thinbox/rip-utils/trever.xmcd";

  `cp \"$cache/$alcat/$aleid\" \"$trever\"`;
}

sub revert_xmcd {
  my $al = shift();
  my $alcat = $al->freedb_cat();
  my $aleid = $al->freedb_eid();
  my $cache = "$ENV{HOME}/.cddb";
  my $trever = "$ENV{HOME}/.thinbox/rip-utils/trever.xmcd";

  `cp \"$trever\" \"$cache/$alcat/$aleid\"`;
}

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
sub prompt {
  my ($question, $default) = @_;
  my $answer;

  print STDOUT "$question [$default] ";
  chop($answer = <STDIN>);
  print STDOUT "\n";

  return $default unless $answer;
  $answer =~ s/^(\s*)(.*?)(\s*)$/$2/; # Remove accidental whitespace
  return $answer;
}

1;